<?php

namespace Drupal\helper;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ExtensionList;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\KeyValueStore\KeyValueFactoryInterface;
use Drupal\Core\State\StateInterface;

/**
 * Helpers related to working with install profiles.
 */
class InstallProfile {

  /**
   * The profile extension list service.
   *
   * @var \Drupal\Core\Extension\ExtensionList
   */
  protected $profileList;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The module schema key/value store.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueStoreInterface
   */
  protected $schemaKeyValue;

  /**
   * InstallProfileSwitcher constructor.
   *
   * @param \Drupal\Core\Extension\ExtensionList $profileList
   *   The profile extension list service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\KeyValueStore\KeyValueFactoryInterface $key_value_factory
   *   The key/value storage factory.
   */
  public function __construct(ExtensionList $profileList, ConfigFactoryInterface $configFactory, StateInterface $state, ModuleHandlerInterface $module_handler, KeyValueFactoryInterface $key_value_factory) {
    $this->profileList = $profileList;
    $this->configFactory = $configFactory;
    $this->state = $state;
    $this->moduleHandler = $module_handler;
    $this->schemaKeyValue = $key_value_factory->get('system.schema');
  }

  /**
   * Switches the site's install profile.
   *
   * Note the following:
   * - This does not run the new profile's install hooks.
   * - Does not enable any of the profile's dependencies.
   * - This does not check that currently enabled modules are located in the
   *   current profile's code and will no longer be available once switching
   *   to the new profile.
   * - Does not run any uninstallation of the current profile, like removing
   *   any tables defined in hook_schema().
   *
   * @param string $profile
   *   The machine name of the profile to switch to.
   * @param int $schema_version
   *   The schema version to set for the install profile.
   *
   * @author https://www.drupal.org/project/profile_switcher
   */
  public function switch($profile, $schema_version = NULL) {
    $this->validateProfile($profile);

    $current_profile = \Drupal::installProfile();
    if ($profile === $current_profile) {
      trigger_error("The current install profile is already set to $profile.", E_USER_NOTICE);
      return;
    }

    // Forces ExtensionDiscovery to rerun for profiles.
    $this->state->delete('system.profile.files');

    // Set the profile in configuration.
    $extension_config = $this->configFactory->getEditable('core.extension');
    $extension_config->set('profile', $profile);
    $extension_config->save();

    drupal_flush_all_caches();

    // Install profiles are also registered as enabled modules.
    // Remove the old profile and add in the new one.
    $extension_config->clear("module.{$current_profile}");
    // The install profile is always given a weight of 1000 by the core
    // extension system.
    $extension_config->set("module.$profile", 1000);
    $extension_config->save();

    // Clear caches again.
    drupal_flush_all_caches();

    // Find the correct schema version of the install profile to set as the
    // current schema version. Assume use the latest schema version if a value
    // has not been provided.
    $schema_version = $schema_version ?: $this->getLatestSchemaVersion($profile);

    // Remove the schema value for the old install profile, and set the schema
    // for the new one.
    $this->schemaKeyValue->delete($current_profile);
    $this->schemaKeyValue->set($profile, $schema_version);

    // Clear caches again.
    drupal_flush_all_caches();
  }

  /**
   * Validates a profile and its hook_requirements() for any errors.
   *
   * @param string $profile
   *   The profile name.
   *
   * @throws \InvalidArgumentException
   *   If the profile to switch to does not exist.
   * @throws \RuntimeException
   *   If the profile failed its hook_requirements().
   */
  public function validateProfile($profile) {
    // Ensure the profile exists.
    if (!$this->profileList->exists($profile)) {
      throw new \InvalidArgumentException("The install profile $profile does not exist.");
    }

    // Make sure the installation API is available.
    include_once DRUPAL_ROOT . '/core/includes/install.inc';

    // Ensure that requirements are checked for the new profile.
    $requirements = drupal_check_profile($profile);
    if (is_array($requirements) && drupal_requirements_severity($requirements) === REQUIREMENT_ERROR) {
      $reasons = [];
      foreach ($requirements as $id => $requirement) {
        if (isset($requirement['severity']) && $requirement['severity'] === REQUIREMENT_ERROR) {
          if (isset($requirement['value']) && $requirement['value']) {
            $reasons[$id] = t('@requirements_message (Currently using @item version @version)', [
              '@requirements_message' => $requirement['description'],
              '@item' => $requirement['title'],
              '@version' => $requirement['value']
            ]);
          }
          else {
            $reasons[$id] = (string) $requirement['description'];
          }
        }
      }
      throw new \RuntimeException("Failed requirements for install profile {$profile}:\n" . implode("\n", $reasons));
    }
  }

  /**
   * Gets the latest schema version for an install profile.
   *
   * @param string $profile
   *   The install profile name.
   *
   * @return int
   *   The schema version.
   */
  protected function getLatestSchemaVersion($profile) {
    module_load_install($profile);
    $versions = drupal_get_schema_versions($profile) ?? [];
    $versions[] = $this->moduleHandler->invoke($profile, 'update_last_removed') ?? \Drupal::CORE_MINIMUM_SCHEMA_VERSION;
    return max($versions);
  }

}
